/*
 * changed a few by hunsoo@sensor.kaist.ac.kr 1997/8/6
 * If any bug exists, the culprit is me
 *
 */
/*
 * Copyright 1988 by Siemens All Rights Reserved
 * 
 * Permission to use, copy, modify and distribute this software is hereby
 * granted, provided that this copyright notice appear in all copies and that
 * the copyright notice appear in supporting documentation.
 * 
 * written 16-17 Mar 1988 by claus gittinger
 * 
 * produce a new .bdf file
 * 
 */

#include <stdio.h>
#include <errno.h>
#include "defs.h"
#include "common.h"

extern int errno;
static void outputchar(FILE * out, struct character * cp);
int	bbox_check(struct character *cp);

void
output(struct fontset *font, char *filename)
{
	FILE   *out;
	int     nchars;
	int     i;
	struct character *cp;

	out = fopen(filename, "w");
	if (out == NULL) {
		err("%s(): cannot open %s for writing\n", __F, filename);
		return;
	}
	fprintf(out, "STARTFONT %d.%d\n", font->f_rev, font->f_subrev);
	/* !! Makes all comments one contiguous block */
	for (i = 0; i < font->f_ncomments; i++) {
		fprintf(out, "%s\n", font->f_comments[i]);
	}
	fprintf(out, "FONT %s\n", font->f_fontname);
	fprintf(out, "SIZE %d %d %d\n",
		font->f_sizes[0], font->f_sizes[1],
		font->f_sizes[2]);
	fprintf(out, "FONTBOUNDINGBOX %d %d %d %d\n",
		font->f_boundingbox.bb_width, font->f_boundingbox.bb_height,
		font->f_boundingbox.bb_offx, font->f_boundingbox.bb_offy);
	fprintf(out, "STARTPROPERTIES %d\n", font->f_nprops);
	for (i = 0; i < font->f_nprops; i++) {
		fprintf(out, "%s\n", font->f_props[i]);
	}

	fprintf(out, "ENDPROPERTIES\n");
	fprintf(out, "CHARS %d\n", font->f_nchars);

	nchars = font->f_nchars;
	cp = font->f_characters[0];

	while (nchars--)
		outputchar(out, cp++);

	fprintf(out, "ENDFONT\n");
	fclose(out);
	return;
}


static void
outputchar(FILE *out, struct character *cp)
{
	int     i, j, k, returned;
	int	start, end, skip0, skip1, howmany;
	unsigned int copied, accum;
	char	*p;

	bbp = &cp->c_bbox;
	fprintf(out, "STARTCHAR %s\n", cp->c_charname);
	if (cp->c_notadobe == 0)
		fprintf(out, "ENCODING %d\n", cp->c_encoding);
	else
		fprintf(out, "ENCODING -1 %d\n", cp->c_encoding);
	fprintf(out, "SWIDTH %d %d\n", cp->c_swidth[0], cp->c_swidth[1]);
	fprintf(out, "DWIDTH %d %d\n", cp->c_dwidth[0], cp->c_dwidth[1]);

	returned = bbox_check(cp);
	fprintf(out, "BBX %d %d %d %d\n",
		bbp->bb_width, bbp->bb_height, bbp->bb_offx, bbp->bb_offy);
	fprintf(out, "BITMAP\n");
	if (returned) {
		fprintf(out, "00\n");
		fprintf(out, "ENDCHAR\n");
		return;
	}

	start = fbbp->bb_height + fbbp->bb_offy - bbp->bb_offy - bbp->bb_height;
	end   = fbbp->bb_height + fbbp->bb_offy - bbp->bb_offy;

	skip0  = bbp->bb_offx - fbbp->bb_offx;
	skip1  = bbp->bb_width;
	howmany = ((skip1 + 3)/4);

	for (i = start; i < end; i++) { /* for */
		copied = cp->c_rows[i];
		for (j=0; j < skip0; j++) {
			if (copied & 0x80000000) {
				err("%s(%s): [%d->%d] : [%d:%d] copied = %8x",
					__F, cp->c_charname,
					start, end, i, j, copied);
			}
			copied <<= 1;
		}
		p = inbuf;
		for (j=0; j < howmany; j++) {
			accum = 0;
			for (k = 0; k <4; k++, copied <<= 1) {
				if (copied & 0x80000000)
					accum += 1;
				accum <<= 1;
			}
			accum >>= 1;
			sprintf(p, "%x", accum);
			p++;
		}
		if ((p-inbuf) % 2)
			*p++ = '0'; /* to make strlen() == even */
		*p++ = 0;
		fprintf(out, "%s\n", inbuf);
	}
	fprintf(out, "ENDCHAR\n");

	return;
}

int
bbox_check(struct character *cp)
{
	int	i, howmany, height;
	unsigned int mask;
	unsigned int max;
	int	 ycnt0, ycnt1;
	unsigned int  *rowbits, bitsum = 0;

	bbp = &cp->c_bbox;
	howmany = 32 - fbbp->bb_width;
	for (mask = 1, i = 0; i < howmany; i++)
		mask <<= 1;

	mask -= 1;
	mask = ~mask;

	bitsum = max = height = 0;
	ycnt0 = ycnt1 = -1;
	for (i = 0; i < fbbp->bb_height; i++) {
		rowbits   = &cp->c_rows[i];
		*rowbits &= mask;
		if (*rowbits == 0)
			continue;
		if (max < *rowbits)
			max = *rowbits;
		if (ycnt0 == -1) {
			ycnt0 = i;
			ycnt1 = i;
		} else
			ycnt1 = i;
		bitsum |= *rowbits;
	}


	bbp->bb_height = ycnt1 - ycnt0 + 1;
	if (ycnt1 == -1) {
		bbp->bb_width = bbp->bb_height = 1;
		bbp->bb_offx = bbp->bb_offy = 0;
		return 1;	/* special case */
	}

	bbp->bb_offy  = fbbp->bb_height + fbbp->bb_offy - ycnt1-1;

	mask = 0x00000001;
	for (i = 0; i < 32 && ((mask & bitsum) == 0); i++, mask <<= 1)
		/* void */;
	ycnt0 = i;
	mask = 0x80000000;
	for (i=32; i > 0 && ((mask & bitsum) == 0); i--, mask >>= 1)
		/* void */;
	ycnt1 = i;

	bbp->bb_width = ycnt1 - ycnt0;
	bbp->bb_offx = fbbp->bb_offx + (32-ycnt1);

	return 0;
}
